/*
===============================================================================
   ___________.___ _______  ________    _________._________________________
   \_   _____/|   |\      \ \______ \  /   _____/|   \__    ___/\_   _____/
    |    __)  |   |/   |   \ |    |  \ \_____  \ |   | |    |    |    __)_
    |     \   |   /    |    \|    `   \/        \|   | |    |    |        \
    \___  /   |___\____|__  /_______  /_______  /|___| |____|   /_______  /
        \/                \/        \/        \/                        \/
                                          __        __
                          ____ ___  ___  / /_____ _/ /
                         / __ `__ \/ _ \/ __/ __ `/ / 
                        / / / / / /  __/ /_/ /_/ / /  
                       /_/ /_/ /_/\___/\__/\__,_/_/   

   FINDSITE-METAL predicts metal-binding sites, residues and metal preferences
   from template structures identified by threading

   Make sure you have libf2c, gzstream, zlib and libsvm installed

   This software is distributed WITHOUT ANY WARRANTY

   Please report bugs to michal@gatech.edu

   Center for the Study of Systems Biology
   250 14th Street NW, Atlanta, GA 30318, USA
===============================================================================
*/


#include "findsitemetal.h"

using namespace std;

int main(int argc, char *argv[])
{
 double      cut_tmscore  = 0.40;
 double      cut_seqid    = 1.00;
 double      cut_tmest    = 1.00;
 double      cut_closepro = 0.90;
 double      cut_closemet = 4.00;
 double      cut_clustdis = 8.00;
 double      cut_binrest  = 0.15;
 int         cut_binresn  = 1;
 std::string tms_norm     = "T";
 
 time_t t_start, t_end;
 time(&t_start);
 srand(time(0));
 
 cout << "------------------------------------------------------------" << endl
      << "                       FINDSITE-METAL" << endl
      << "               metal-binding site prediction" << endl << endl
      << "          Please report bugs to michal@gatech.edu" << endl
      << "------------------------------------------------------------" << endl << endl;
 
 if ( argc < 7 )
 {
  cout << " findsitemetal -s <target structure, PDB>" << endl
       << "               -t <list of templates>" << endl
       << "               -o <output filename>" << endl << endl
       << " additional options:" << endl
       << "          -b <sequence identity threshold (default 1.0)>" << endl
       << "          -c <estimated TMscore (default 1.0)>" << endl
       << "          -m <TMscore threshold (default 0.4)>" << endl
       << "          -r <SVM threshold (default 0.15)>" << endl
       << "          -n <min # of binding residues (default 1)>" << endl
       << "          -l <TMscore normalization: (default T)>" << endl
       << "              T - target length" << endl
       << "              S - smaller protein" << endl << endl;
  
  exit(EXIT_SUCCESS);
 }
 
 string target_name;
 bool target_opt = false;
 
 string templates_name;
 bool templates_opt = false;
 
 string output_name;
 bool output_opt = false;
 
 for ( int i = 0; i < argc; i++ )
 {
  if ( !strcmp(argv[i],"-s") && i < argc ) { target_name    = string(argv[i+1]); target_opt    = true; }
  if ( !strcmp(argv[i],"-t") && i < argc ) { templates_name = string(argv[i+1]); templates_opt = true; }
  if ( !strcmp(argv[i],"-o") && i < argc ) { output_name    = string(argv[i+1]); output_opt    = true; }
  if ( !strcmp(argv[i],"-b") && i < argc ) { cut_seqid      = atof(argv[i+1]);                         }
  if ( !strcmp(argv[i],"-c") && i < argc ) { cut_tmest      = atof(argv[i+1]);                         }
  if ( !strcmp(argv[i],"-m") && i < argc ) { cut_tmscore    = atof(argv[i+1]);                         }
  if ( !strcmp(argv[i],"-r") && i < argc ) { cut_binrest    = atof(argv[i+1]);                         }
  if ( !strcmp(argv[i],"-n") && i < argc ) { cut_binresn    = atoi(argv[i+1]);                         }
  if ( !strcmp(argv[i],"-l") && i < argc ) { tms_norm       = string(argv[i+1]);                       }
 }
 
 char * path1;
 
 path1 = getenv("FINDSITELIB"); if ( path1==NULL ) { cout << "FINDSITELIB is not set" << endl; exit(EXIT_FAILURE); }
 
 path1 = getenv("FINDSITEMAP"); if ( path1==NULL ) { cout << "FINDSITEMAP is not set" << endl; exit(EXIT_FAILURE); }
 
 path1 = getenv("FINDSITEDAT"); if ( path1==NULL ) { cout << "FINDSITEDAT is not set" << endl; exit(EXIT_FAILURE); }
 
 string lib_path;
 lib_path = getenv("FINDSITELIB");
 
 string cluster_def;
 cluster_def = getenv("FINDSITEMAP");
 
 string svm_stuff;
 svm_stuff = getenv("FINDSITEDAT");
 
 struct svm_model * fs_model;
 
 struct svm_node * fs_node;
 
 int fs_attr = 64;
 
 fs_node = ( struct svm_node * ) malloc( fs_attr * sizeof( struct svm_node ) );
 
 double fs_scale[MAXSVM][2];
 
 if ( ( fs_model = svm_load_model( (svm_stuff+"/findsitemetalSVM.model").c_str() ) ) == 0 )
 { cout << "Cannot open " << (svm_stuff+"/findsitemetalSVM.model").c_str() << endl; exit(EXIT_FAILURE); }
 
 if ( SVMscaling( (svm_stuff+"/findsitemetalSVM.scale").c_str(), fs_scale ) )
 { cout << "Cannot open " << (svm_stuff+"/findsitemetalSVM.scale").c_str() << endl; exit(EXIT_FAILURE); }
 
 if ( !target_opt )
 {
  cout << "Please provide target structure" << endl;
  exit(EXIT_FAILURE);
 }
 
 if ( !templates_opt )
 {
  cout << "Please provide list of templates" << endl;
  exit(EXIT_FAILURE);
 }
 
 if ( !output_opt )
 {
  cout << "Please provide output filename" << endl;
  exit(EXIT_FAILURE);
 }
 else
 {
  ofstream outsit1( (output_name+".sites.dat").c_str() );
  outsit1.close();
  
  ofstream outsit2( (output_name+".sites.pdb").c_str() );
  outsit2.close();
  
  ofstream outsit3( (output_name+".metals.pdb").c_str() );
  outsit3.close();
  
  ofstream outsit4( (output_name+".templates.pdb").c_str() );
  outsit4.close();
  
  ofstream outsit5( (output_name+".alignments.dat").c_str() );
  outsit5.close();
 }
 
 if ( tms_norm != "T" && tms_norm != "S" )
 {
  cout << "TM-score can be normalized either by the target length (T) or the lenght of the smaller protein (S)" << endl;
  exit(EXIT_FAILURE);
 }
 
 if ( cut_seqid < 1.0 )
  cout << "Benchmark mode (" << cut_seqid << ")" << endl << endl;
 
 Template * target;
 
 target = new Template( 0, 0, 0 );
 
 bool load0 = target->loadProtein(target_name, false);
 
 if ( load0 )
 {
  cout << "Cannot read target structure" << endl;
  exit(EXIT_FAILURE);
 }
 
 list<string> complex_list1;
 
 getList( templates_name, lib_path, cluster_def, complex_list1 );
 
 list< Template * > template_set;
 
 list<string>::iterator it1;
 
 for ( it1 = complex_list1.begin(); it1 != complex_list1.end(); it1++ )
 {
  bool load1, load2, load3;
  
  Template * template_tmp = new Template( 0, 0, 0 );
  
  load1 = template_tmp->loadProtein(lib_path+"/PROTEINS/"+(*it1).substr(1,2)+"/"+(*it1), true);
  
  load2 = template_tmp->loadMetal(lib_path+"/METALS/"+(*it1).substr(1,2)+"/"+(*it1), true);
  
  load3 = template_tmp->filterMetal(cut_closemet);
  
  template_tmp->setProteinID(*it1);
  
  if ( !load1 && !load2 && !load3 )
   template_set.push_back( template_tmp );
  else
   delete template_tmp;
 }
 
 if ( template_set.empty() )
 {
  cout << "No metal-bound structure templates have been identified" << endl << endl;
  
  template_set.clear();
  
  time(&t_end);
  
  printTime( difftime(t_end, t_start) );
  
  exit(EXIT_SUCCESS);
 }
 
 cout << endl << "Template filtering ... " << flush;
 
 list< Template * > template_set_filtered;
 
 list< Template * >::iterator tpl1;
 list< Template * >::iterator tpl2;
 
 for ( tpl1 = template_set.begin(); tpl1 != template_set.end(); tpl1++ )
 {
  double sid1 = (*tpl1)->alignNW( target->getProteinSequence(), true );
  
  if ( sid1 <= cut_seqid )
  {
   double tms1 = (*tpl1)->alignTM( target, true, tms_norm );
   
   if ( tms1 >= cut_tmscore )
   {
    bool tpl3 = true;
    
    for ( tpl2 = template_set_filtered.begin(); tpl2 != template_set_filtered.end(); tpl2++ )
    {
     double sid2 = (*tpl1)->alignNW( (*tpl2)->getProteinSequence(), false );
     
     if ( sid2 > cut_closepro )
     {
      binding_metal   bin_m1[MAXMET];
      binding_residue bin_r1[MAXMET][MAXRES];
      
      binding_metal   bin_m2[MAXMET];
      binding_residue bin_r2[MAXMET][MAXRES];
      
      int bin_n1 = (*tpl1)->getMetal( bin_m1, bin_r1 );
      int bin_n2 = (*tpl2)->getMetal( bin_m2, bin_r2 );
      
      for ( int bin_i1 = 0; bin_i1 < bin_n1; bin_i1++ )
      {
       bool bin_w1 = true;
       
       for ( int bin_i2 = 0; bin_i2 < bin_n2; bin_i2++ )
        if ( bin_m1[bin_i1].metal_type == bin_m2[bin_i2].metal_type )
         if ( sqrt(pow(bin_m1[bin_i1].metal_xyz[0]-bin_m2[bin_i2].metal_xyz[0], 2) + pow(bin_m1[bin_i1].metal_xyz[1]-bin_m2[bin_i2].metal_xyz[1], 2) + pow(bin_m1[bin_i1].metal_xyz[2]-bin_m2[bin_i2].metal_xyz[2], 2)) < cut_closemet )
          bin_w1 = false;
       
       if ( !bin_w1 )
       {
        for ( int bin_i3 = bin_i1 + 1; bin_i3 < bin_n1; bin_i3++ )
        {
         bin_m1[bin_i3-1].metal_name     = bin_m1[bin_i3].metal_name;
         bin_m1[bin_i3-1].metal_type     = bin_m1[bin_i3].metal_type;
         bin_m1[bin_i3-1].metal_residues = bin_m1[bin_i3].metal_residues;
         bin_m1[bin_i3-1].metal_site     = bin_m1[bin_i3].metal_site;
         
         for ( int bin_i4 = 0; bin_i4 < 3; bin_i4++ )
          bin_m1[bin_i3-1].metal_xyz[bin_i4] = bin_m1[bin_i3].metal_xyz[bin_i4];
  
         for ( int bin_i5 = 0; bin_i5 < bin_m1[bin_i3].metal_residues; bin_i5++ )
         {
          bin_r1[bin_i3-1][bin_i5].residue_number = bin_r1[bin_i3][bin_i5].residue_number;
          bin_r1[bin_i3-1][bin_i5].residue_name   = bin_r1[bin_i3][bin_i5].residue_name;
         }
        }
        
        bin_i1--;
        bin_n1--;
       }
      }
      
      if ( bin_n1 < 1 )
       tpl3 = false;
      else
       (*tpl1)->setMetal( bin_m1, bin_r1, bin_n1 );
     }
     
     if ( !tpl3 )
      break;
    }
    
    if ( tpl3 )
     template_set_filtered.push_back( *tpl1 );
   }
  }
 }
 
 template_set.clear();
 
 if ( template_set_filtered.empty() )
 {
  cout << "no metal-bound templates survived" << endl << endl;
  
  template_set.clear();
  template_set_filtered.clear();
  
  time(&t_end);
  
  printTime( difftime(t_end, t_start) );
  
  exit(EXIT_SUCCESS);
 }
 
 cout << template_set_filtered.size() << " metal-bound templates survived" << endl << endl;
 
 cout << "Detecting metal-binding sites ... " << flush;
 
 double sim1[MAXSIM][3];
 
 int tpl1s = 0;
 
 for ( tpl1 = template_set_filtered.begin(); tpl1 != template_set_filtered.end(); tpl1++ )
 {
  binding_metal   bin_m1[MAXMET];
  binding_residue bin_r1[MAXMET][MAXRES];
  
  int bin_n1 = (*tpl1)->getMetal( bin_m1, bin_r1 );
  
  for ( int bin_i1 = 0; bin_i1 < bin_n1; bin_i1++ )
  {
   for ( int bin_i2 = 0; bin_i2 < 3; bin_i2++ )
    sim1[tpl1s][bin_i2] = bin_m1[bin_i1].metal_xyz[bin_i2];
   
   tpl1s++;
  }
 }
 
 double (*sim2)[MAXSIM] = new double[MAXSIM][MAXSIM];
 
 for ( int tpl1i = 0; tpl1i < tpl1s - 1; tpl1i++ )
  for ( int tpl2i = tpl1i + 1; tpl2i < tpl1s; tpl2i++ )
   sim2[tpl1i][tpl2i] = sqrt( pow( sim1[tpl1i][0] - sim1[tpl2i][0], 2 ) + 
                              pow( sim1[tpl1i][1] - sim1[tpl2i][1], 2 ) + 
                              pow( sim1[tpl1i][2] - sim1[tpl2i][2], 2 ) );
 
 int * clu1 = new int [MAXSIM];
 
 int clu2 = cluster_avelink( sim2, clu1, tpl1s, cut_clustdis, "min" );
 
 delete [] sim2;
 
 if ( clu2 < 1 )
 {
  cout << "no metal-binding sites found" << endl << endl;
  
  template_set.clear();
  template_set_filtered.clear();
  
  time(&t_end);
  
  printTime( difftime(t_end, t_start) );
  
  exit(EXIT_SUCCESS);
 }
 
 cout << clu2 << " metal-binding sites found" << endl << endl;
 
 int tpl2s = 0;
 
 for ( tpl1 = template_set_filtered.begin(); tpl1 != template_set_filtered.end(); tpl1++ )
 {
  binding_metal   bin_m2[MAXMET];
  binding_residue bin_r2[MAXMET][MAXRES];
  
  int bin_n2 = (*tpl1)->getMetal( bin_m2, bin_r2 );
  
  for ( int bin_i2 = 0; bin_i2 < bin_n2; bin_i2++ )
   bin_m2[bin_i2].metal_site = clu1[tpl2s++] + 1;
  
  (*tpl1)->setMetal( bin_m2, bin_r2, bin_n2 );
 }
 
 list< Site * > site_set;
 
 for ( int clu3 = 0; clu3 < clu2; clu3++ )
 {
  Site * site_tmp = new Site( clu3 + 1 );
  
  for ( tpl1 = template_set_filtered.begin(); tpl1 != template_set_filtered.end(); tpl1++ )
  {
   binding_metal   bin_m3[MAXMET];
   binding_residue bin_r3[MAXMET][MAXRES];
   
   int bin_n3 = (*tpl1)->getMetal( bin_m3, bin_r3 );
   
   int bin_s3 = 0;
   
   for ( int bin_i3 = 0; bin_i3 < bin_n3; bin_i3++ )
    if ( bin_m3[bin_i3].metal_site == clu3 +1 )
     bin_s3++;
   
   if ( bin_s3 > 0 )
   {
    list<string> complex_list2;
    
    site_tmp->getListProteins( complex_list2 );
    
    list<string>::iterator complex_pos2;
    
    bool bin_w3 = true;
    
    for ( complex_pos2 = complex_list2.begin(); complex_pos2 != complex_list2.end(); complex_pos2++ )
     if ( (*complex_pos2) == (*tpl1)->getProteinID() )
      bin_w3 = false;
    
    if ( bin_w3 )
     site_tmp->addTemplate(*tpl1);
    
    site_tmp->setMetalsTotal( site_tmp->getMetalsTotal() + bin_s3 );
   }
  }
  
  if ( site_tmp->getProteinsTotal() > 0 && site_tmp->getMetalsTotal() > 0 )
   site_set.push_back( site_tmp );
  else
   delete site_tmp;
 }
 
 template_set.clear();
 
 delete [] clu1;
 
 cout << "Calculating binding residues ... " << flush;
 
 list< Site * > site_set_filtered;
 
 list< Site * >::iterator isit1;
 
 for ( isit1 = site_set.begin(); isit1 != site_set.end(); isit1++ )
 {
  (*isit1)->calculateSiteCenter();
  
  int bres1 = (*isit1)->calculateBindingResidues( target->getProteinResiduesTotal(), (target->getProteinSequence()).c_str(), target );
  
  if ( bres1 >= cut_binresn )
   site_set_filtered.push_back( *isit1 );
 }
 
 site_set.clear();
 
 if ( site_set_filtered.empty() )
 {
  cout << "no sites survived" << endl << endl;
  
  template_set.clear();
  template_set_filtered.clear();
  
  site_set_filtered.clear();
  
  time(&t_end);
  
  printTime( difftime(t_end, t_start) );
  
  exit(EXIT_SUCCESS);
 }
 
 cout << site_set_filtered.size() << " sites survived" << endl << endl;
 
 cout << "Calculating metal preferences ... " << flush;
 
 int totmet1 = 0;
 
 for ( isit1 = site_set_filtered.begin(); isit1 != site_set_filtered.end(); isit1++ )
 {
  (*isit1)->calculateMetalPrefs();
  (*isit1)->calculateTMscore();
  
  totmet1 += (*isit1)->getMetalsTotal();
 }
 
 for ( isit1 = site_set_filtered.begin(); isit1 != site_set_filtered.end(); isit1++ )
  (*isit1)->setSiteFraction( ( (double) (*isit1)->getMetalsTotal() ) / ( (double) totmet1 ) );
 
 cout << "done" << endl << endl;
 
 cout << "Applying SVM ... " << flush;
 
 list< Site * > site_set_filtered_svm;
 
 for ( isit1 = site_set_filtered.begin(); isit1 != site_set_filtered.end(); isit1++ )
 {
  int bres2 = (*isit1)->applySVM( fs_model, fs_node, fs_scale, cut_tmest, cut_binrest );
  
  if ( bres2 >= cut_binresn )
   site_set_filtered_svm.push_back( *isit1 );
 }
 
 site_set_filtered.clear();
 
 if ( site_set_filtered_svm.empty() )
 {
  cout << "no sites survived" << endl << endl;
  
  template_set.clear();
  template_set_filtered.clear();
  
  site_set_filtered_svm.clear();
  
  time(&t_end);
  
  printTime( difftime(t_end, t_start) );
  
  exit(EXIT_SUCCESS);
 }
 
 cout << site_set_filtered_svm.size() << " sites survived" << endl << endl;
 
 multimap<double, Site *> site_set_sorted;
 
 for ( isit1 = site_set_filtered_svm.begin(); isit1 != site_set_filtered_svm.end(); isit1++ )
  site_set_sorted.insert( pair<double, Site *>( -1.0 * (*isit1)->getSiteScore( 1 ), (*isit1) ) );
 
 site_set_filtered_svm.clear();
 
 map<string,bool> chk1;
 
 for ( tpl1 = template_set_filtered.begin(); tpl1 != template_set_filtered.end(); tpl1++ )
  if ( chk1.count( (*tpl1)->getProteinID() ) < 1 )
  {
   chk1[(*tpl1)->getProteinID()] = true;
   
   double tarca1[MAXPRO][3];
   
   target->getProteinCoordsCA(tarca1);
   
   (*tpl1)->dumpProtein( output_name );
   (*tpl1)->dumpAlignment( output_name, target->getProteinResiduesTotal(), target->getProteinSequence(), tarca1 );
  }
 
 int atnum1 = 1;
 int stnum1 = 1;
 
 multimap<double, Site *>::iterator isit2;
 
 for ( isit2 = site_set_sorted.begin() ; isit2 != site_set_sorted.end(); isit2++ )
 {
  atnum1 = ((*isit2).second)->dumpMetals( output_name, atnum1, stnum1 );
  
  ((*isit2).second)->dumpSite( output_name, target, cut_binrest, stnum1 );
  
  stnum1++;
 }
 
 template_set.clear();
 template_set_filtered.clear();
 
 site_set_sorted.clear();
 
 time(&t_end);
 
 printTime( difftime(t_end, t_start) );
 
 return 0;
}
