/*
===============================================================================
   ___________.___ _______  ________    _________._________________________
   \_   _____/|   |\      \ \______ \  /   _____/|   \__    ___/\_   _____/
    |    __)  |   |/   |   \ |    |  \ \_____  \ |   | |    |    |    __)_
    |     \   |   /    |    \|    `   \/        \|   | |    |    |        \
    \___  /   |___\____|__  /_______  /_______  /|___| |____|   /_______  /
        \/                \/        \/        \/                        \/
                                          __        __
                          ____ ___  ___  / /_____ _/ /
                         / __ `__ \/ _ \/ __/ __ `/ / 
                        / / / / / /  __/ /_/ /_/ / /  
                       /_/ /_/ /_/\___/\__/\__,_/_/   

   FINDSITE-METAL predicts metal-binding sites, residues and metal preferences
   from template structures identified by threading

   Make sure you have libf2c, gzstream, zlib and libsvm installed

   This software is distributed WITHOUT ANY WARRANTY

   Please report bugs to michal@gatech.edu

   Center for the Study of Systems Biology
   250 14th Street NW, Atlanta, GA 30318, USA
===============================================================================
*/


#include "site.h"

using namespace std;

Site::Site( int an )
{
 _site_number = an;
 
 _site_fraction = 0.0;
 _site_TMave    = 0.0;
 _site_TMstd    = 0.0;
 
 _seq_score1 = 0.0;
 _seq_score2 = 0.0;
 _seq_score3 = 0;
 
 _metal_tot = 0;
 
 _metal_best1 = 0;
 _metal_best2 = -1.0;
 
 _complexes.clear();
 _metal_res.clear();
 
 for ( int i1 = 0; i1 < 6; i1++ )
  _site_center[i1] = 0.0;
}

Site::Site( void )
{
 _site_number = 0;
 
 _site_fraction = 0.0;
 _site_TMave    = 0.0;
 _site_TMstd    = 0.0;
 
 _seq_score1 = 0.0;
 _seq_score2 = 0.0;
 _seq_score3 = 0;
 
 _metal_tot = 0;
 
 _metal_best1 = 0;
 _metal_best2 = -1.0;
 
 _complexes.clear();
 _metal_res.clear();
 
 for ( int i1 = 0; i1 < 6; i1++ )
  _site_center[i1] = 0.0;
}

Site::~Site() {}

// ==================================================================================   addTemplate

void Site::addTemplate( Template * template_new1 )
{
 _complexes.push_back( template_new1 );
 
 _list_prot.clear();
 
 for ( list< Template * >::iterator tp1 = _complexes.begin(); tp1 != _complexes.end(); tp1++ )
  _list_prot.push_back( (*tp1)->getProteinID() );
 
 _list_prot.sort();
 _list_prot.unique();
}

// ==================================================================================   getProteinsTotal

int Site::getProteinsTotal( void )
{
 return _list_prot.size();
}

// ==================================================================================   getMetalsTotal

int Site::getMetalsTotal( void )
{
 return _metal_tot;
}

// ==================================================================================   setMetalsTotal

void Site::setMetalsTotal( int mt1 )
{
 _metal_tot = mt1;
}

// ==================================================================================   getSiteFraction

double Site::getSiteFraction( void )
{
 return _site_fraction;
}

// ==================================================================================   getSiteScore

double Site::getSiteScore( int sc1 )
{
 switch ( sc1 )
 {
  case  1  :  return _seq_score1;
  case  2  :  return _seq_score2;
  case  3  :  return _seq_score3;
  
  default  :  return 0;
 }
}

// ==================================================================================   setSiteFraction

void Site::setSiteFraction( double mt1 )
{
 _site_fraction = mt1;
}

// ==================================================================================   setSiteNumber

void Site::setSiteNumber( int mt1 )
{
 _site_number = mt1;
}

// ==================================================================================   getSiteNumber

int Site::getSiteNumber( void )
{
 return _site_number;
}

// ==================================================================================   getListProteins

void Site::getListProteins( list<string> &pkt_list1 )
{
 pkt_list1 = _list_prot;
}

// ==================================================================================   calculateSiteCenter

void Site::calculateSiteCenter( void )
{
 list< Template * >::iterator ipt1;
 
 for ( int ipt2 = 0; ipt2 < 6; ipt2++ )
  _site_center[ipt2] = 0.0;
 
 for ( ipt1 = _complexes.begin(); ipt1 != _complexes.end(); ipt1++ )
 {
  binding_metal   s_m1[MAXMET];
  binding_residue s_r1[MAXMET][MAXRES];
  
  int s_n1 = (*ipt1)->getMetal( s_m1, s_r1 );
  
  for ( int s_i1 = 0; s_i1 < s_n1; s_i1++ )
   if ( s_m1[s_i1].metal_site == _site_number )
    for ( int ipt2 = 0; ipt2 < 3; ipt2++ )
    {
     _site_center[ipt2] += s_m1[s_i1].metal_xyz[ipt2];
     
     _site_center[ipt2+3] += pow( s_m1[s_i1].metal_xyz[ipt2], 2.0 );
    }
 }
 
 if ( _metal_tot > 1 )
  for ( int ipt2 = 0; ipt2 < 3; ipt2++ )
   _site_center[ipt2+3] = sqrt( ( (double) _metal_tot * _site_center[ipt2+3] - _site_center[ipt2] * _site_center[ipt2] ) / ( (double) _metal_tot * ( (double) _metal_tot - 1.0 ) ) );
 else
  for ( int ipt2 = 3; ipt2 < 6; ipt2++ )
  _site_center[ipt2] = 0.0;
 
 for ( int ipt2 = 0; ipt2 < 3; ipt2++ )
  _site_center[ipt2] /= (double) _metal_tot;
}

// ==================================================================================   calculateBindingResidues

int Site::calculateBindingResidues( int res1, std::string tseq1, Template * target_str1 )
{
 _metal_res.clear();
 
 list< Template * >::iterator ipt1;
 
 map<int,metal_residue>::iterator ipt4;
 
 for ( ipt1 = _complexes.begin(); ipt1 != _complexes.end(); ipt1++ )
 {
  int aln1[MAXPRO];
  
  (*ipt1)->getTMalignment(aln1, res1);
  
  binding_metal   s_m1[MAXMET];
  binding_residue s_r1[MAXMET][MAXRES];
  
  int s_n1 = (*ipt1)->getMetal( s_m1, s_r1 );
  
  for ( int ipt2 = 0; ipt2 < res1; ipt2++ )
   if ( aln1[ipt2] > -1 )
    for ( int s_i1 = 0; s_i1 < s_n1; s_i1++ )
     if ( s_m1[s_i1].metal_site == _site_number )
      for ( int s_i2 = 0; s_i2 < s_m1[s_i1].metal_residues; s_i2++ )
       if ( aln1[ipt2] == s_r1[s_i1][s_i2].residue_number )
       {
        ipt4 = _metal_res.find( ipt2 );
        
        if ( ipt4 != _metal_res.end() )
        {
         ((*ipt4).second).template_incontact++;
         
         ((*ipt4).second).residue_profile1[one2num(s_r1[s_i1][s_i2].residue_name)] += 1.0;
         ((*ipt4).second).residue_profile2[one2num(s_r1[s_i1][s_i2].residue_name)] += 1.0;
        }
        else
        {
         metal_residue res4;
         
         res4.residue_number = ipt2;
         res4.residue_code = ( tseq1.substr( ipt2, 1 ) ).c_str();
         res4.template_incontact = 1;
         res4.template_fraction = 0.0;
         res4.template_fraction_bkg = 0.0;
         res4.residue_distance = 0.0;
         res4.residue_svm = 0.0;
         
         for ( int s_i3 = 0; s_i3 < 20; s_i3++ )
         {
          res4.residue_profile1[s_i3] = 0.0;
          res4.residue_profile2[s_i3] = 0.0;
         }
         
         res4.residue_profile1[one2num(s_r1[s_i1][s_i2].residue_name)] = 1.0;
         res4.residue_profile2[one2num(s_r1[s_i1][s_i2].residue_name)] = 1.0;
         
         _metal_res.insert( pair<int,metal_residue>(ipt2,res4) );
        }
       }
 }
 
 double target_str2[MAXPRO][3];
 
 target_str1->getProteinCoordsCA(target_str2);
 
 for ( ipt4 = _metal_res.begin() ; ipt4 != _metal_res.end(); ipt4++ )
 {
  if ( _metal_tot > 0 && ((*ipt4).second).template_incontact > 0 )
  {
   ((*ipt4).second).template_fraction = ( (double) ( ((*ipt4).second).template_incontact ) ) / ( (double) _metal_tot );
   
   ((*ipt4).second).template_fraction_bkg = ( ( (double) ( ((*ipt4).second).template_incontact ) ) + getBkgFreq1( tseq1.substr( ((*ipt4).second).residue_number, 1) ) * sqrt( (double) _metal_tot ) ) / ( ( (double) _metal_tot ) + sqrt( (double) _metal_tot ) );
   
   ((*ipt4).second).residue_distance = sqrt( pow( target_str2[((*ipt4).second).residue_number][0] - _site_center[0], 2 ) + pow( target_str2[((*ipt4).second).residue_number][1] - _site_center[1], 2 ) + pow( target_str2[((*ipt4).second).residue_number][2] - _site_center[2], 2 ) );
   
   ((*ipt4).second).residue_svm = 0.0;
   
   for ( int s_i4 = 0; s_i4 < 20; s_i4++ )
   {
    ((*ipt4).second).residue_profile1[s_i4] /= ( (double) ( ((*ipt4).second).template_incontact ) );
    
    ((*ipt4).second).residue_profile2[s_i4] = ( ((*ipt4).second).residue_profile2[s_i4] + getBkgFreq1( num2one(s_i4) ) * sqrt( (double) ( ((*ipt4).second).template_incontact ) ) ) / ( ( (double) ( ((*ipt4).second).template_incontact ) ) + sqrt( (double) ( ((*ipt4).second).template_incontact ) ) );
   }
  }
  else
   _metal_res.erase(ipt4--);
 }
 
 return _metal_res.size();
}

// ==================================================================================   calculateMetalPrefs

void Site::calculateMetalPrefs( void )
{
 for ( int ipm1 = 0; ipm1 < 8; ipm1++ )
  for ( int ipm2 = 0; ipm2 < 2; ipm2++ )
   _metal_pref[ipm1][ipm2] = 0.0;
 
 for ( int ipm1 = 0; ipm1 < 3; ipm1++ )
  _metal_entropy[ipm1] = 0.0;
 
 list< Template * >::iterator ipt1;
 
 for ( ipt1 = _complexes.begin(); ipt1 != _complexes.end(); ipt1++ )
 {
  binding_metal   s_m1[MAXMET];
  binding_residue s_r1[MAXMET][MAXRES];
  
  int s_n1 = (*ipt1)->getMetal( s_m1, s_r1 );
  
  for ( int s_i1 = 0; s_i1 < s_n1; s_i1++ )
   if ( s_m1[s_i1].metal_site == _site_number )
    for ( int s_i2 = 0; s_i2 < 2; s_i2++ )
     _metal_pref[s_m1[s_i1].metal_type][s_i2] += 1.0;
 }
 
 _metal_best1 = 0;
 _metal_best2 = -1.0;
 
 for ( int ipm1 = 0; ipm1 < 8; ipm1++ )
 {
  _metal_pref[ipm1][0] /= _metal_tot;
  
  _metal_pref[ipm1][1] = ( _metal_pref[ipm1][1] + getBkgFreq2( ipm1 ) * sqrt( (double) _metal_tot ) ) / ( (double) _metal_tot + sqrt( (double) _metal_tot ) );
  
  if ( _metal_pref[ipm1][0] > 0.0 )
   _metal_entropy[0] -= _metal_pref[ipm1][0] * log2(_metal_pref[ipm1][0]);
  
  if ( _metal_pref[ipm1][1] > 0.0 )
   _metal_entropy[1] -= _metal_pref[ipm1][1] * log2(_metal_pref[ipm1][1]);
  
  if ( getBkgFreq2( ipm1 ) > 0.0 )
   _metal_entropy[2] -= getBkgFreq2( ipm1 ) * log2(getBkgFreq2( ipm1 ));
  
  if ( _metal_pref[ipm1][1] > _metal_best2 )
  {
   _metal_best1 = ipm1;
   _metal_best2 = _metal_pref[ipm1][1];
  }
 }
 
 for ( int ipm1 = 0; ipm1 < 3; ipm1++ )
  _metal_entropy[ipm1] = abs( _metal_entropy[ipm1] );
}

// ==================================================================================   calculateTMscore

void Site::calculateTMscore( void )
{
 _site_TMave = 0.0;
 _site_TMstd = 0.0;
 
 list< Template * >::iterator ipt1;
 
 for ( ipt1 = _complexes.begin(); ipt1 != _complexes.end(); ipt1++ )
 {
  _site_TMave += (*ipt1)->getProteinTMscore();
  _site_TMstd += pow( (*ipt1)->getProteinTMscore(), 2.0 );
 }
 
 if ( _complexes.size() > 1 )
  _site_TMstd = sqrt( ( (double) _complexes.size() * _site_TMstd - _site_TMave * _site_TMave ) / ( (double) _complexes.size() * ( (double) _complexes.size() - 1.0 ) ) );
 else
  _site_TMstd = 0.0;
 
 _site_TMave /= (double) _complexes.size();
}

// ==================================================================================   dumpMetals

int Site::dumpMetals( std::string m1_name, int m1_number, int m2_number )
{
 list< Template * >::iterator ipt1;
 
 ofstream outmet( (m1_name+".metals.pdb").c_str(), ios_base::out|ios_base::app );
 
 for ( ipt1 = _complexes.begin(); ipt1 != _complexes.end(); ipt1++ )
 {
  binding_metal   d_m1[MAXMET];
  binding_residue d_r1[MAXMET][MAXRES];
  
  int d_n1 = (*ipt1)->getMetal( d_m1, d_r1 );
  
  for ( int d_i1 = 0; d_i1 < d_n1; d_i1++ )
   if ( d_m1[d_i1].metal_site == _site_number )
    outmet << "HETATM" << setw(5) << m1_number++
                       << setw(3) << d_m1[d_i1].metal_name
                       << setw(6) << d_m1[d_i1].metal_name
                       << setw(6) << m2_number
                       << fixed << setw(12) << setprecision(3) << d_m1[d_i1].metal_xyz[0]
                       << fixed << setw(8) << setprecision(3) << d_m1[d_i1].metal_xyz[1]
                       << fixed << setw(8) << setprecision(3) << d_m1[d_i1].metal_xyz[2] << endl;
 }
 
 outmet << "TER" << endl;
 
 outmet.close();
 
 return m1_number;
}

// ==================================================================================   dumpSite

void Site::dumpSite( std::string p1_name, Template * target_str3, double tres1, int s1_number )
{
 ofstream outsit1( (p1_name+".sites.pdb").c_str(), ios_base::out|ios_base::app );
 ofstream outsit2( (p1_name+".sites.dat").c_str(), ios_base::out|ios_base::app );
 
 list< Template * >::iterator itpl1;
 
 outsit1 << "HETATM" << setw(5) << s1_number
                     << "  FS  PKT M"
                     << setw(4) << _metal_tot
                     << fixed << setw(12) << setprecision(3) << _site_center[0]
                     << fixed << setw(8) << setprecision(3) << _site_center[1]
                     << fixed << setw(8) << setprecision(3) << _site_center[2] << endl;
 
 outsit2 << "SITE" << setw(7) << s1_number
                   << fixed << setw(5) << _metal_tot
                   << fixed << setw(8) << setprecision(4) << _site_fraction
                   << fixed << setw(8) << setprecision(4) << _seq_score1
                   << fixed << setw(8) << setprecision(4) << _site_TMave
                   << fixed << setw(8) << setprecision(4) << _site_TMstd
                   << fixed << setw(5) << _seq_score3 << endl;
 
 map<string,bool> chk2;
 
 int itpl3 = 0;
 
 for ( itpl1 = _complexes.begin(); itpl1 != _complexes.end(); itpl1++ )
  if ( chk2.count( (*itpl1)->getProteinID() ) < 1 )
  {
   chk2[(*itpl1)->getProteinID()] = true;
   
   outsit2 << "TEMPLATE" << setw(5) << ++itpl3 << setw(6) << (*itpl1)->getProteinID() 
                         << fixed << setw(5) << setprecision(0) << (*itpl1)->getProteinResiduesTotal()
                         << fixed << setw(5) << setprecision(0) << (*itpl1)->getProteinLengthTM()
                         << fixed << setw(8) << setprecision(3) << (*itpl1)->getProteinSeqID1()
                         << fixed << setw(8) << setprecision(3) << (*itpl1)->getProteinSeqID2()
                         << fixed << setw(8) << setprecision(3) << (*itpl1)->getProteinTMscore()
                         << fixed << setw(8) << setprecision(3) << (*itpl1)->getProteinRMSD() << endl;
  }
 
 for ( int itpl4 = 0; itpl4 < 8; itpl4++ )
  outsit2 << "METAL" << fixed << setw(3) << num2two(itpl4)
                     << fixed << setw(9) << setprecision(6) << _metal_pref[itpl4][0]
                     << fixed << setw(9) << setprecision(6) << _metal_pref[itpl4][1] << endl;
 
 outsit2 << "MENTROPY" << fixed << setw(9) << setprecision(3) << _metal_entropy[0]
                       << fixed << setw(9) << setprecision(3) << _metal_entropy[1]
                       << fixed << setw(9) << setprecision(3) << _metal_entropy[2] << endl;
 
 outsit2 << "CENTER" << fixed << setw(9) << setprecision(3) << _site_center[0]
                     << fixed << setw(9) << setprecision(3) << _site_center[1]
                     << fixed << setw(9) << setprecision(3) << _site_center[2]
                     << fixed << setw(9) << setprecision(3) << _site_center[3]
                     << fixed << setw(9) << setprecision(3) << _site_center[4]
                     << fixed << setw(9) << setprecision(3) << _site_center[5] << endl;
 
 map<int,metal_residue>::iterator itpl5;
 
 for ( itpl5 = _metal_res.begin() ; itpl5 != _metal_res.end(); itpl5++ )
 {
  string bind1 = " ";
  
  if ( ((*itpl5).second).residue_svm >= tres1 )
   bind1 = "*";
  
  outsit2 << "RESIDUE" << setw(6) << target_str3->getProteinNumbering(((*itpl5).second).residue_number)
                       << setw(2) << ((*itpl5).second).residue_code
                       << setw(2) << bind1
                       << setw(5) << ((*itpl5).second).template_incontact
                       << fixed << setw(8) << setprecision(3) << ((*itpl5).second).residue_distance
                       << fixed << setw(9) << setprecision(5) << ((*itpl5).second).residue_svm
                       << fixed << setw(8) << setprecision(5) << ((*itpl5).second).template_fraction_bkg
                       << fixed << setw(8) << setprecision(5) << ((*itpl5).second).residue_profile1[one2num( ((*itpl5).second).residue_code[0] )]
                       << fixed << setw(8) << setprecision(5) << ((*itpl5).second).residue_profile2[one2num( ((*itpl5).second).residue_code[0] )]
                       << fixed << setw(8) << setprecision(5) << getBkgFreq3( _metal_best1, one2num( ((*itpl5).second).residue_code[0] ) ) << endl;
 }
 
 outsit2 << "TER" << endl;
}

// ==================================================================================   applySVM

int Site::applySVM( struct svm_model * & fsvm1, struct svm_node * & fsvm2, double fsvm3[][2], double fsvm4, double fsvm5 )
{
 _seq_score1 = 0.0;
 _seq_score2 = 0.0;
 _seq_score3 = 0;
 
 double fsvm6[MAXSVM];
 
 map<int,metal_residue>::iterator itpl1;
 
 for ( itpl1 = _metal_res.begin() ; itpl1 != _metal_res.end(); itpl1++ )
 {
  fsvm6[0] = fsvm4;
  fsvm6[1] = ((*itpl1).second).residue_distance;
  fsvm6[2] = ((*itpl1).second).template_fraction;
  fsvm6[3] = ((*itpl1).second).template_fraction_bkg;
  fsvm6[4] = _site_TMave;
  fsvm6[5] = _site_TMstd;
  fsvm6[6] = ( _site_center[3] + _site_center[4] + _site_center[5] ) / 3.0;
  fsvm6[7] = _metal_entropy[1];
  fsvm6[8] = ((*itpl1).second).residue_profile2[one2num( ((*itpl1).second).residue_code[0] )];
  
  for ( int itpl2 = 0; itpl2 < 8; itpl2++ )
   fsvm6[itpl2+9] = _metal_pref[itpl2][1];
  
  for ( int itpl2 = 0; itpl2 < 8; itpl2++ )
   fsvm6[itpl2+17] = getBkgFreq3( itpl2, one2num( ((*itpl1).second).residue_code[0] ) );
  
  for ( int itpl3 = 0; itpl3 < MAXSVM; itpl3++ )
   SVMnormalize( fsvm6[itpl3], fsvm3[itpl3][0], fsvm3[itpl3][1] );
  
  ((*itpl1).second).residue_svm = SVMpredict( fsvm1, fsvm2, fsvm6 );
  
  if ( ((*itpl1).second).residue_svm >= fsvm5 )
  {
   _seq_score1 += ((*itpl1).second).residue_profile2[one2num( ((*itpl1).second).residue_code[0] )];
   
   if ( ((*itpl1).second).residue_profile2[one2num( ((*itpl1).second).residue_code[0] )] != 0.0 )
    _seq_score2 -= ((*itpl1).second).residue_profile2[one2num( ((*itpl1).second).residue_code[0] )] * log2( ((*itpl1).second).residue_profile2[one2num( ((*itpl1).second).residue_code[0] )] );
   
   _seq_score3++;
  }
 }
 
 return _seq_score3;
}
