/*
===============================================================================
   ___________.___ _______  ________    _________._________________________
   \_   _____/|   |\      \ \______ \  /   _____/|   \__    ___/\_   _____/
    |    __)  |   |/   |   \ |    |  \ \_____  \ |   | |    |    |    __)_
    |     \   |   /    |    \|    `   \/        \|   | |    |    |        \
    \___  /   |___\____|__  /_______  /_______  /|___| |____|   /_______  /
        \/                \/        \/        \/                        \/
                                          __        __
                          ____ ___  ___  / /_____ _/ /
                         / __ `__ \/ _ \/ __/ __ `/ / 
                        / / / / / /  __/ /_/ /_/ / /  
                       /_/ /_/ /_/\___/\__/\__,_/_/   

   FINDSITE-METAL predicts metal-binding sites, residues and metal preferences
   from template structures identified by threading

   Make sure you have libf2c, gzstream, zlib and libsvm installed

   This software is distributed WITHOUT ANY WARRANTY

   Please report bugs to michal@gatech.edu

   Center for the Study of Systems Biology
   250 14th Street NW, Atlanta, GA 30318, USA
===============================================================================
*/


#ifndef __SITE_H_
#define __SITE_H_

#include<list>
#include<fstream>
#include<iostream>
#include<iomanip>
#include<string>
#include<bitset>
#include<map>

#include "size.h"
#include "template.h"
#include "data.h"
#include "runsvm.h"

using namespace std;

struct metal_residue
{
 int         residue_number;
 std::string residue_code;
 int         template_incontact;
 double      template_fraction;
 double      template_fraction_bkg;
 double      residue_profile1[20];
 double      residue_profile2[20];
 double      residue_distance;
 double      residue_svm;
};

class Site {
        
  private:
    
    list< Template * >           _complexes;            // template complexes
    
    double                       _site_fraction;        // number of metals-bound template
    double                       _site_TMave;           // average TM-score
    double                       _site_TMstd;           // standard dev TMscore
    
    int                          _site_number;          // site number
    int                          _metal_tot;            // number of metals
    
    int                          _metal_best1;          // predicted metal
    double                       _metal_best2;          // predicted metal
    
    double                       _site_center[6];       // site geometric center ( ave, sd )
    
    list<std::string>            _list_prot;            // list of proteins
    
    map<int,metal_residue>       _metal_res;            // metal-binding residues
    
    double                       _metal_pref[8][2];     // binding metal preferences
    
    double                       _metal_entropy[3];     // binding metal entropy
    
    double                       _seq_score1;           // sequence score 1
    double                       _seq_score2;           // sequence score 2
    int                          _seq_score3;           // sequence score 3
    
  public:
    
    Site( int );
    
    Site( void );
    
    ~Site();
    
    void addTemplate( Template * );
    
    int getProteinsTotal( void );
    
    int getMetalsTotal( void );
    
    void setMetalsTotal( int );
    
    double getSiteFraction( void );
    
    double getSiteScore( int );
    
    void setSiteFraction( double  );
    
    int getSiteNumber( void );
    
    void setSiteNumber( int );
    
    void getListProteins( list<string> & );
    
    void calculateSiteCenter( void );
    
    int calculateBindingResidues( int, std::string, Template * );
    
    void calculateMetalPrefs( void );
    
    void calculateTMscore( void );
    
    int dumpMetals( std::string, int, int );
    
    void dumpSite( std::string, Template *, double, int );
    
    int applySVM( struct svm_model * &, struct svm_node * &, double [][2], double, double );
};

#endif
