/*
===============================================================================
   ___________.___ _______  ________    _________._________________________
   \_   _____/|   |\      \ \______ \  /   _____/|   \__    ___/\_   _____/
    |    __)  |   |/   |   \ |    |  \ \_____  \ |   | |    |    |    __)_
    |     \   |   /    |    \|    `   \/        \|   | |    |    |        \
    \___  /   |___\____|__  /_______  /_______  /|___| |____|   /_______  /
        \/                \/        \/        \/                        \/
                                          __        __
                          ____ ___  ___  / /_____ _/ /
                         / __ `__ \/ _ \/ __/ __ `/ / 
                        / / / / / /  __/ /_/ /_/ / /  
                       /_/ /_/ /_/\___/\__/\__,_/_/   

   FINDSITE-METAL predicts metal-binding sites, residues and metal preferences
   from template structures identified by threading

   Make sure you have libf2c, gzstream, zlib and libsvm installed

   This software is distributed WITHOUT ANY WARRANTY

   Please report bugs to michal@gatech.edu

   Center for the Study of Systems Biology
   250 14th Street NW, Atlanta, GA 30318, USA
===============================================================================
*/


#include "template.h"

using namespace std;

Template::Template( int aa, int ar, int am )
{
 _na      = aa;
 _nr      = ar;
 _nm      = am;
 
 _score_nw  = 0.0;
 _score_tm  = 0.0;
 _score_rms = 0.0;
 _score_id  = 0.0;
 
 _score_len    = 0;
 _score_len_nw = 0;
}

Template::Template( void )
{
 _na      = 0;
 _nr      = 0;
 _nm      = 0;
 
 _score_nw  = 0.0;
 _score_tm  = 0.0;
 _score_rms = 0.0;
 _score_id  = 0.0;
 
 _score_len    = 0;
 _score_len_nw = 0;
}

Template::~Template() {}

// ==================================================================================   setProteinID

void Template::setProteinID( string pid1 )
{
 _protein_id = pid1;
}

// ==================================================================================   getProteinID

string Template::getProteinID( void )
{
 return _protein_id;
}

// ==================================================================================   getProteinSeqID1

double Template::getProteinSeqID1( void )
{
 return _score_nw;
}

// ==================================================================================   getProteinSeqID2

double Template::getProteinSeqID2( void )
{
 return _score_id;
}

// ==================================================================================   getProteinTMscore

double Template::getProteinTMscore( void )
{
 return _score_tm;
}

// ==================================================================================   getProteinRMSD

double Template::getProteinRMSD( void )
{
 return _score_rms;
}

// ==================================================================================   getProteinLengthNW

int Template::getProteinLengthNW( void )
{
 return _score_len_nw;
}

// ==================================================================================   getProteinLengthTM

int Template::getProteinLengthTM( void )
{
 return _score_len;
}

// ==================================================================================   loadProtein

bool Template::loadProtein( std::string p1_name, bool p1_gzip )
{
 string atom1, residue1;
 
 int atom2, residue2;
 
 double x1, y1, z1;
 
 list<string> p1_data;
 list<string>::iterator p1_i;
 
 if ( p1_gzip )
 {
  string line1;
  
  igzstream p1_file( p1_name.c_str() );
  
  if ( ! p1_file.good() )
  {
   std::cerr << "ERROR: Opening file `" << p1_name << "' failed\n";
   return EXIT_FAILURE;
  }
  
  while ( getline(p1_file,line1) )
   p1_data.push_back( line1 );
   
  p1_file.close();
 }
 
 else
 {
  string line1;
  
  ifstream p1_file( p1_name.c_str() );
  
  if ( !p1_file.is_open() )  { cout << "Cannot open " << p1_name << endl; exit(EXIT_FAILURE); }
  
  while (getline(p1_file,line1))
   p1_data.push_back( line1 );
  
  p1_file.close();
 }
 
 for ( p1_i = p1_data.begin(); p1_i != p1_data.end(); p1_i++ )
  if ( (*p1_i).size() > 53 )
   if ( (*p1_i).substr(0,6) == "ATOM  ")
   {
    residue1 = (*p1_i).substr(17,3);
    atom1    = (*p1_i).substr(12,4);
    
    residue2 = atoi((*p1_i).substr(22,4).c_str());
    atom2    = atoi((*p1_i).substr(6,5).c_str());
    
    x1       = atof((*p1_i).substr(30,8).c_str());
    y1       = atof((*p1_i).substr(38,8).c_str());
    z1       = atof((*p1_i).substr(46,8).c_str());
    
    _protein_xyz.push_back( CoordsProtein( atom2, residue2, x1, y1, z1, residue1, atom1 ) );
    
    _na++;
    
    if ( atom1 == " CA " )
    {
     _protein_seq1.append( three2oneS( (*p1_i).substr(17,3) ) );
     
     _protein_seq3[_nr] = residue2;
     
     _nr++;
    }
   }
 
 strcpy(_protein_seq2, _protein_seq1.c_str());
 
 return EXIT_SUCCESS;
}

// ==================================================================================   loadMetal

bool Template::loadMetal( std::string m1_name, bool m1_gzip )
{
 list<string> m1_data;
 list<string>::iterator m1_i;
 
 if ( m1_gzip )
 {
  string line1;
  
  igzstream m1_file( m1_name.c_str() );
  
  if ( ! m1_file.good() )
  {
   std::cerr << "ERROR: Opening file `" << m1_name << "' failed\n";
   return EXIT_FAILURE;
  }
  
  while ( getline(m1_file,line1) )
   m1_data.push_back( line1 );
   
  m1_file.close();
 }
 
 else
 {
  string line1;
  
  ifstream m1_file( m1_name.c_str() );
  
  if ( !m1_file.is_open() )  { cout << "Cannot open " << m1_name << endl; exit(EXIT_FAILURE); }
  
  while ( getline(m1_file,line1) )
   m1_data.push_back( line1 );
  
  m1_file.close();
 }
 
 for ( m1_i = m1_data.begin(); m1_i != m1_data.end(); m1_i++ )
 {
  string metal1 = (*m1_i).substr(0,2);
  
  _binding_m[_nm].metal_name = metal1;
  _binding_m[_nm].metal_type = two2num(metal1);
  
  _binding_m[_nm].metal_xyz[0] = atof((*m1_i).substr(2,9).c_str());
  _binding_m[_nm].metal_xyz[1] = atof((*m1_i).substr(11,9).c_str());
  _binding_m[_nm].metal_xyz[2] = atof((*m1_i).substr(20,9).c_str());
  
  _binding_m[_nm].metal_residues = atoi((*m1_i).substr(29,3).c_str());
  
  _binding_m[_nm].metal_site = 0;
  
  string line2;
  
  int m2_i = 0;
  
  istringstream iss((*m1_i).substr(33,(*m1_i).size()-33), istringstream::in);
  
  while( iss >> line2 )     
  {
   _binding_r[_nm][m2_i].residue_number = atoi(line2.substr(0,line2.size()-2).c_str()) - 1;
   _binding_r[_nm][m2_i].residue_name = line2[line2.size()-1];
   
   m2_i++;
  }
  
  _nm++;
 }
 
 return EXIT_SUCCESS;
}

// ==================================================================================   filterMetal

bool Template::filterMetal( double mthres1 )
{
 if ( _nm > 1 )
 {
  double (*msim1)[MAXMET] = new double[MAXMET][MAXMET];
  
  for ( int mi1 = 0; mi1 < _nm - 1; mi1++ )
   for ( int mi2 = mi1 + 1; mi2 < _nm; mi2++ )
    msim1[mi1][mi2] = sqrt( pow( _binding_m[mi1].metal_xyz[0] - _binding_m[mi2].metal_xyz[0], 2 ) + 
                            pow( _binding_m[mi1].metal_xyz[1] - _binding_m[mi2].metal_xyz[1], 2 ) + 
                            pow( _binding_m[mi1].metal_xyz[2] - _binding_m[mi2].metal_xyz[2], 2 ) );
  
  int * mclu1 = new int [MAXMET];
  
  int mclu2 = cluster_avelink_metal( msim1, mclu1, _nm, mthres1, "min" );
  
  delete [] msim1;
  
  binding_metal   f_m1[MAXMET];
  binding_residue f_r1[MAXMET][MAXRES];
  int             f_n1 = 0;
  
  for ( int mclu3 = 0; mclu3 < mclu2; mclu3++ )
  {
   int bst1 = 0;
   double bst2 = 1e6;
   
   for ( int mi3 = 0; mi3 < _nm; mi3++ )
    if ( mclu1[mi3] == mclu3 )
    {
     double ave1 = 0.0;
     
     for ( int mi4 = 0; mi4 < _nm; mi4++ )
      if ( mclu1[mi4] == mclu3 && mi4 != mi3 )
       ave1 += sqrt( pow( _binding_m[mi3].metal_xyz[0] - _binding_m[mi4].metal_xyz[0], 2 ) + 
                     pow( _binding_m[mi3].metal_xyz[1] - _binding_m[mi4].metal_xyz[1], 2 ) + 
                     pow( _binding_m[mi3].metal_xyz[2] - _binding_m[mi4].metal_xyz[2], 2 ) );
     
     if ( ave1 < bst2 )
     {
      bst1 = mi3;
      bst2 = ave1;
     }
    }
   
   f_m1[f_n1].metal_name = _binding_m[bst1].metal_name;
   f_m1[f_n1].metal_type = _binding_m[bst1].metal_type;
   
   for ( int mi5 = 0; mi5 < 3; mi5++ )
    f_m1[f_n1].metal_xyz[mi5] = _binding_m[bst1].metal_xyz[mi5];
   
   f_m1[f_n1].metal_residues = _binding_m[bst1].metal_residues;
   f_m1[f_n1].metal_site = _binding_m[bst1].metal_site;
   
   for ( int mi6 = 0; mi6 < _binding_m[bst1].metal_residues; mi6++ )
   {
    f_r1[f_n1][mi6].residue_number = _binding_r[bst1][mi6].residue_number;
    
    f_r1[f_n1][mi6].residue_name = _binding_r[bst1][mi6].residue_name;
   }
   
   f_n1++;
  }
  
  for ( int mclu4 = 0; mclu4 < f_n1; mclu4++ )
  {
   _binding_m[mclu4].metal_name = f_m1[mclu4].metal_name;
   _binding_m[mclu4].metal_type = f_m1[mclu4].metal_type;
   
   for ( int mi6 = 0; mi6 < 3; mi6++ )
    _binding_m[mclu4].metal_xyz[mi6] = f_m1[mclu4].metal_xyz[mi6];
   
   _binding_m[mclu4].metal_residues = f_m1[mclu4].metal_residues;
   _binding_m[mclu4].metal_site = f_m1[mclu4].metal_site;
   
   for ( int mi7 = 0; mi7 < f_m1[mclu4].metal_residues; mi7++ )
   {
    _binding_r[mclu4][mi7].residue_number = f_r1[mclu4][mi7].residue_number;
    
    _binding_r[mclu4][mi7].residue_name = f_r1[mclu4][mi7].residue_name;
   }
  }
  
  _nm = f_n1;
 }
 
 if ( _nm > 0 )
  return EXIT_SUCCESS;
 else
  return EXIT_FAILURE;
}

// ==================================================================================   getProteinSequence

std::string Template::getProteinSequence( void )
{
 return _protein_seq1;
}

// ==================================================================================   alignNW

double Template::alignNW( std::string target_seq1, bool target_ctr1 )
{
 double target_sco1, target_sco2;
 
 int target_len1 = _protein_seq1.size();
 int target_len2 = target_seq1.size();
 
 char target_seq2[MAXPRO];
 
 strcpy(target_seq2, target_seq1.c_str());
 
 int target_ali1[MAXPRO];
 
 nwalign_( &target_sco1, &target_sco2, target_len1, target_len2, _protein_seq2, target_seq2, target_ali1 );
 
 if ( target_ctr1 )
 {
  for ( int ia1 = 0; ia1 < target_len2; ia1++ )
   _protein_align1[ia1] = target_ali1[ia1];
  
  _score_nw     = target_sco1;
  _score_len_nw = target_len2;
 }
 
 return target_sco1;
}

// ==================================================================================   alignTM

double Template::alignTM( Template * target_str1, bool target_ctr1, std::string target_ctr2 )
{
 int    target_sco1; // alignment length
 double target_sco2; // RMSD of the aligned residues
 double target_sco3; // TM-score
 double target_sco4; // sid of the aligned region
 
 target_sco1 = 0;
 
 double p_xyz1a[MAXPRO][3];
 double p_xyz1b[MAXPRO][3];
 
 int p_len1a = 0;
 int p_len1b = 0;
 
 int p_res1a[MAXPRO];
 int p_res1b[MAXPRO];
 
 double tt1[3];
 double uu1[3][3];
 
 int align1[MAXPRO];
 
 vector<CoordsProtein>::iterator ip1;
 vector<CoordsProtein>::iterator ip2;
 
 for ( ip1 = _protein_xyz.begin(); ip1 < _protein_xyz.end(); ip1++ )
  if ( (*ip1).getAtomName() == " CA " )
  {
   for ( int ip3 = 0; ip3 < 3; ip3++ )
    p_xyz1a[p_len1a][ip3] = (*ip1).getCoords(ip3+1);
   
   p_res1a[p_len1a] = p_len1a + 1;
   
   p_len1a++;
  }
 
 for ( p_len1b = 0; p_len1b < target_str1->getProteinResiduesTotal(); p_len1b++ )
  p_res1b[p_len1b] = p_len1b + 1;
 
 p_len1b = target_str1->getProteinCoordsCA(p_xyz1b);
 
 char target_seq2[MAXPRO];
 
 strcpy(target_seq2, (target_str1->getProteinSequence()).c_str());
 
 int lnorm = p_len1b;
 
 if ( target_ctr2 == "S" && p_len1a < p_len1b )
  lnorm = p_len1a;
 
 frtmalign_( &target_sco1, &target_sco2, &target_sco3, &target_sco4, p_len1a, p_len1b, _protein_seq2, target_seq2, align1, p_res1a, p_res1b, p_xyz1a, p_xyz1b, tt1, uu1, lnorm );
 
 if ( target_sco1 )
 {
  target_sco4 = 0.0;
  
  for ( int ii2 = 0; ii2 < p_len1b; ii2++ )
  {
   if ( align1[ii2] > -1 )
    if ( target_seq2[ii2] == _protein_seq2[align1[ii2]] )
     target_sco4 += 1.0;
  }
  
  target_sco4 = target_sco4 / ( (double) target_sco1 );
 }
 
 if ( target_ctr1 )
 {
  for ( int ii3 = 0; ii3 < p_len1b; ii3++ )
   _protein_align2[ii3] = align1[ii3];
  
  for ( ip2 = _protein_xyz.begin(); ip2 < _protein_xyz.end(); ip2++ )
  {
   double xyz_t1[3];
   
   for ( int ip3 = 0; ip3 < 3; ip3++ )
    xyz_t1[ip3] = (*ip2).getCoords(ip3+1);
   
   for ( int ip3 = 0; ip3 < 3; ip3++ )
    xyz_t1[ip3] = tt1[ip3] + uu1[ip3][0] * (*ip2).getCoords(1) + uu1[ip3][1] * (*ip2).getCoords(2) + uu1[ip3][2] * (*ip2).getCoords(3);
   
   (*ip2).setCoords(xyz_t1[0], xyz_t1[1], xyz_t1[2]);
  }
  
  for ( int im1 = 0; im1 < _nm; im1++ )
  {
   double xyz_t1[3];
   
   for ( int im2 = 0; im2 < 3; im2++ )
    xyz_t1[im2] = _binding_m[im1].metal_xyz[im2];
   
   for ( int im2 = 0; im2 < 3; im2++ )
    _binding_m[im1].metal_xyz[im2] = tt1[im2] + uu1[im2][0] * xyz_t1[0] + uu1[im2][1] * xyz_t1[1] + uu1[im2][2] * xyz_t1[2];
  }
  
  _score_tm  = target_sco3;
  _score_rms = target_sco2;
  _score_id  = target_sco4;
  _score_len = target_sco1;
 }
 
 return target_sco3;
}

// ==================================================================================   getNWalignment

void Template::getNWalignment( int tab1[], int tab2 )
{
 for ( int ial1 = 0; ial1 < tab2; ial1++ )
  tab1[ial1] = _protein_align1[ial1];
}

// ==================================================================================   getTMalignment

void Template::getTMalignment( int tab1[], int tab2 )
{
 for ( int ial1 = 0; ial1 < tab2; ial1++ )
  tab1[ial1] = _protein_align2[ial1];
}

// ==================================================================================   getProteinResiduesTotal

int Template::getProteinResiduesTotal( void )
{
 return _nr;
}

// ==================================================================================   getProteinCoordsCA

int Template::getProteinCoordsCA( double tabCA1[][3] )
{
 vector<CoordsProtein>::iterator ipca1;
 
 int ipca2 = 0;
 
 for ( ipca1 = _protein_xyz.begin(); ipca1 < _protein_xyz.end(); ipca1++ )
  if ( (*ipca1).getAtomName() == " CA " )
  {
   for ( int ipca3 = 0; ipca3 < 3; ipca3++ )
    tabCA1[ipca2][ipca3] = (*ipca1).getCoords(ipca3+1);
   
   ipca2++;
  }
 
 return ipca2;
}

// ==================================================================================   getMetal

int Template::getMetal( binding_metal m1[], binding_residue r1[][MAXRES] )
{
 for ( int im1 = 0; im1 < _nm; im1++ )
 {
  m1[im1].metal_name     = _binding_m[im1].metal_name;
  m1[im1].metal_type     = _binding_m[im1].metal_type;
  m1[im1].metal_residues = _binding_m[im1].metal_residues;
  m1[im1].metal_site     = _binding_m[im1].metal_site;
  
  for ( int im2 = 0; im2 < 3; im2++ )
   m1[im1].metal_xyz[im2] = _binding_m[im1].metal_xyz[im2];
  
  for ( int im3 = 0; im3 < _binding_m[im1].metal_residues; im3++ )
  {
   r1[im1][im3].residue_number = _binding_r[im1][im3].residue_number;
   r1[im1][im3].residue_name   = _binding_r[im1][im3].residue_name;
  }
 }
 
 return _nm;
}

// ==================================================================================   setMetal

void Template::setMetal( binding_metal m1[], binding_residue r1[][MAXRES], int n1 )
{
 for ( int im1 = 0; im1 < n1; im1++ )
 {
  _binding_m[im1].metal_name     = m1[im1].metal_name;
  _binding_m[im1].metal_type     = m1[im1].metal_type;
  _binding_m[im1].metal_residues = m1[im1].metal_residues;
  _binding_m[im1].metal_site     = m1[im1].metal_site;
  
  for ( int im2 = 0; im2 < 3; im2++ )
   _binding_m[im1].metal_xyz[im2] = m1[im1].metal_xyz[im2];
  
  for ( int im3 = 0; im3 < m1[im1].metal_residues; im3++ )
  {
   _binding_r[im1][im3].residue_number = r1[im1][im3].residue_number;
   _binding_r[im1][im3].residue_name = r1[im1][im3].residue_name;
  }
 }
 
 _nm = n1;
}

// ==================================================================================   dumpProtein

void Template::dumpProtein( std::string p1_name )
{
 vector<CoordsProtein>::iterator p1_i;
 
 ofstream outprot( (p1_name+".templates.pdb").c_str(), ios_base::out|ios_base::app );
 
 outprot << "REMARK   PDB-ID:  " << setw(8) << _protein_id << endl
         << "REMARK   SEQID1:  " << fixed << setw(8) << setprecision(3) << _score_nw << endl
         << "REMARK   SEQID2:  " << fixed << setw(8) << setprecision(3) << _score_id << endl
         << "REMARK   TM-SCORE:" << fixed << setw(8) << setprecision(3) << _score_tm << endl
         << "REMARK   RMSD:    " << fixed << setw(8) << setprecision(3) << _score_rms << endl;
 
 for ( p1_i = _protein_xyz.begin(); p1_i < _protein_xyz.end(); p1_i++ )
  outprot << "ATOM  " << setw(5) << (*p1_i).getAtomNumber()
                      << setw(5) << (*p1_i).getAtomName()
                      << setw(4) << (*p1_i).getResidueName()
                      << setw(6) << (*p1_i).getResidueNumber()
                      << fixed << setw(12) << setprecision(3) << (*p1_i).getCoords(1)
                      << fixed << setw(8)  << setprecision(3) << (*p1_i).getCoords(2)
                      << fixed << setw(8)  << setprecision(3) << (*p1_i).getCoords(3) << endl;
 
 outprot << "TER" << endl;
 
 outprot.close();
}

// ==================================================================================   dumpAlignment

void Template::dumpAlignment( std::string p1_name, int tarlen1, std::string tarseq1, double tarca2[][3] )
{
 list<int> frsq1;
 list<int> frsq2;
 
 list<int>::iterator isq1;
 list<int>::iterator isq2;
 
 std::string asq1 = "";
 std::string asq2 = "";
 std::string asq3 = "";
 
 for ( int iaf1 = 0; iaf1 < tarlen1; iaf1++ )
 {
  if ( iaf1 == 0 )
  {
   if ( _protein_align2[iaf1] > 0 )
   {
    for ( int iaf2 = 0; iaf2 < _protein_align2[iaf1]; iaf2++ )
    {
     frsq1.push_back(-1);
     frsq2.push_back(iaf2);
    }
   }
   else if ( _protein_align2[iaf1] < 0 )
   {
    int iaf3 = -1;
    
    for ( int iaf2 = 0; iaf2 < tarlen1; iaf2++ )
    {
     if ( _protein_align2[iaf2] > -1 )
     {
      iaf3 = _protein_align2[iaf2];
      
      break;
     }
    }
    
    for ( int iaf2 = 0; iaf2 < iaf3; iaf2++ )
    {
     frsq1.push_back(-1);
     frsq2.push_back(iaf2);
    }
   }
   
   frsq1.push_back(iaf1);
   frsq2.push_back(_protein_align2[iaf1]);
  }
  else if ( iaf1 == tarlen1 -1 )
  {
   frsq1.push_back(iaf1);
   frsq2.push_back(_protein_align2[iaf1]);
   
   if ( _protein_align2[iaf1] >= 0 )
   {
    if ( _protein_align2[iaf1] + 1 < _nr )
    {
     for ( int iaf2 = _protein_align2[iaf1] + 1; iaf2 < _nr; iaf2++ )
     {
      frsq1.push_back(-1);
      frsq2.push_back(iaf2);
     }
    }
   }
   else
   {
    int iaf3 = -1;
    
    for ( int iaf2 = iaf1; iaf2 >= 0 ; iaf2-- )
    {
     if ( _protein_align2[iaf2] > -1 )
     {
      iaf3 = _protein_align2[iaf2];
      
      break;
     }
    }
    
    for ( int iaf2 = 1; iaf2 < _nr - iaf3; iaf2++ )
    {
     frsq1.push_back(-1);
     frsq2.push_back(iaf3+iaf2);
    }
   }
  }
  else
  {
   frsq1.push_back(iaf1);
   frsq2.push_back(_protein_align2[iaf1]);
   
   if ( _protein_align2[iaf1+1] < 0 && _protein_align2[iaf1] > -1 )
   {
    int iaf3 = -1;
    
    for ( int iaf2 = iaf1 + 1; iaf2 < tarlen1; iaf2++ )
    {
     if ( _protein_align2[iaf2] > -1 )
     {
      iaf3 = _protein_align2[iaf2];
      
      break;
     }
    }
    
    for ( int iaf2 = 1; iaf2 < iaf3 - _protein_align2[iaf1]; iaf2++ )
    {
     frsq1.push_back(-1);
     frsq2.push_back(_protein_align2[iaf1] + iaf2);
    }
   }
   else if ( _protein_align2[iaf1] + 1 < _protein_align2[iaf1+1] && _protein_align2[iaf1] > -1 )
   {
    for ( int iaf2 = 1; iaf2 < ( _protein_align2[iaf1+1] - _protein_align2[iaf1] ); iaf2++ )
    {
     frsq1.push_back(-1);
     frsq2.push_back(_protein_align2[iaf1] + iaf2);
    }
   }
  }
 }
 
 double tplca2[MAXPRO][3];
 
 vector<CoordsProtein>::iterator itca1;
 
 int itca2 = 0;
 
 for ( itca1 = _protein_xyz.begin(); itca1 < _protein_xyz.end(); itca1++ )
  if ( (*itca1).getAtomName() == " CA " )
  {
   for ( int itca3 = 0; itca3 < 3; itca3++ )
    tplca2[itca2][itca3] = (*itca1).getCoords(itca3+1);
   
   itca2++;
  }
 
 for ( isq1 = frsq1.begin(), isq2 = frsq2.begin(); isq1 != frsq1.end(); isq1++, isq2++ )
 {
  if ( *isq1 < 0 )
   asq1 += "-";
  else
   asq1 += tarseq1.substr(*isq1,1).c_str();
  
  if ( *isq2 < 0 )
   asq2 += "-";
  else
   asq2 += _protein_seq1.substr(*isq2,1).c_str();
  
  if ( *isq1 > -1 && *isq2 > -1 )
  {
   if ( sqrt( pow(tarca2[*isq1][0]-tplca2[*isq2][0], 2) + pow(tarca2[*isq1][1]-tplca2[*isq2][1], 2) + pow(tarca2[*isq1][2]-tplca2[*isq2][2], 2)) < 5.0 )
    asq3 += ":";
   else
    asq3 += ".";
  }
  else
   asq3 += " ";
 }
 
 ofstream outali( (p1_name+".alignments.dat").c_str(), ios_base::out|ios_base::app );
 
 outali << ">"  << _protein_id << " " << _nr << " " << _score_len << " " << fixed << setprecision(3) << _score_tm << " " << _score_rms << " " << _score_id << endl << asq1 << endl << asq3 << endl << asq2 << endl << "*" << endl;
 
 outali.close();
}

// ==================================================================================   getProteinNumbering

int Template::getProteinNumbering ( int mapp1 )
{
 return _protein_seq3[mapp1];
}
