/*
===============================================================================
   ___________.___ _______  ________    _________._________________________
   \_   _____/|   |\      \ \______ \  /   _____/|   \__    ___/\_   _____/
    |    __)  |   |/   |   \ |    |  \ \_____  \ |   | |    |    |    __)_
    |     \   |   /    |    \|    `   \/        \|   | |    |    |        \
    \___  /   |___\____|__  /_______  /_______  /|___| |____|   /_______  /
        \/                \/        \/        \/                        \/
                                          __        __
                          ____ ___  ___  / /_____ _/ /
                         / __ `__ \/ _ \/ __/ __ `/ / 
                        / / / / / /  __/ /_/ /_/ / /  
                       /_/ /_/ /_/\___/\__/\__,_/_/   

   FINDSITE-METAL predicts metal-binding sites, residues and metal preferences
   from template structures identified by threading

   Make sure you have libf2c, gzstream, zlib and libsvm installed

   This software is distributed WITHOUT ANY WARRANTY

   Please report bugs to michal@gatech.edu

   Center for the Study of Systems Biology
   250 14th Street NW, Atlanta, GA 30318, USA
===============================================================================
*/


#ifndef __TEMPLATE_H_
#define __TEMPLATE_H_

#include<vector>
#include<fstream>
#include<iostream>
#include<iomanip>
#include<cstring>
#include<sstream>
#include<bitset>
#include<list>
#include<cmath>
#include<algorithm>
#include<gzstream.h>

#include "size.h"
#include "coords.h"
#include "nwalign.h"
#include "frtmalign.h"
#include "data.h"
#include "cluster.h"

using namespace std;

struct binding_residue
{
 int    residue_number;
 char   residue_name;
};

struct binding_metal
{
 std::string metal_name;
 int         metal_type;
 double      metal_xyz[3];
 int         metal_residues;
 int         metal_site;
};

class Template {
        
  private:
    
    vector<CoordsProtein>      _protein_xyz;              // protein coordinates
    
    string                     _protein_id;               // protein PDB-ID
    
    int                        _na;                       // number of protein atoms
    int                        _nr;                       // number of protein residues
    int                        _nm;                       // number of bound metals
    
    string                     _protein_seq1;              // aa sequence
    char                       _protein_seq2[MAXPRO];      // aa sequence
    int                        _protein_seq3[MAXPRO];      // aa sequence numbering
    
    int                        _protein_align1[MAXPRO];    // alignment to target from NW
    int                        _protein_align2[MAXPRO];    // alignment to target from TMalign
    
    double                     _score_nw;                  // sequence identity to target
    double                     _score_tm;                  // TM-score to target
    double                     _score_rms;                 // RMSD to target
    double                     _score_id;                  // sequence identity to target (from structure alignment)
    int                        _score_len;                 // structure alignment length
    int                        _score_len_nw;              // sequence alignment length
    
    binding_metal              _binding_m[MAXMET];         // bound metals
    binding_residue            _binding_r[MAXMET][MAXRES]; // binding residues
    
    int                        _pocket_number;             // pocket number
    int                        _pocket_cluster_number;     // metal cluster number
    
  public:
    
    Template( int, int, int );
    
    Template( void );
    
    ~Template();
    
    void setProteinID( std::string );
    
    std::string getProteinID( void );
    
    double getProteinSeqID1( void );
    
    double getProteinSeqID2( void );
    
    double getProteinTMscore( void );
    
    double getProteinRMSD( void );
    
    int getProteinLengthNW( void );
    
    int getProteinLengthTM( void );
    
    bool loadProtein( std::string, bool );
    
    bool loadMetal( std::string, bool );
    
    bool filterMetal( double );
    
    std::string getProteinSequence( void );
    
    double alignNW( std::string, bool );
    
    double alignTM( Template *, bool, std::string );
    
    void getNWalignment( int [], int );
    
    void getTMalignment( int [], int );
    
    int getProteinResiduesTotal( void );
    
    int getProteinCoordsCA( double [][3] );
    
    int getMetal( binding_metal [], binding_residue [][MAXRES] );
    
    void setMetal( binding_metal [], binding_residue [][MAXRES], int );
    
    void dumpProtein( std::string );
    
    void dumpAlignment( std::string, int, std::string, double [][3] );
    
    int getProteinNumbering( int );
};

#endif
